(**************************************************************)
(*   The Strong Completeness of Partial Monoidal Boolean BI   *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.

Require Import predicate.
Require Import closure.
Require Import graph.

Require Import words.
Require Import constraints.
Require Import pme.

Section Elem_simpl.

  Set Implicit Arguments.

  Variable L : Type.

  Variable   basic_cst : (constraint L -> Prop) -> constraint L -> Prop.
  Hypothesis basic_id  : forall R R' c, R ~ R' -> basic_cst R c -> basic_cst R' c.
  Hypothesis basic_eps : forall R, is_pme R -> basic_cst R (eps~~eps).
  
  Inductive basic_pme : (constraint L -> Prop) -> Prop :=
    | basic_pme_empty : forall R, R ~ pme empty -> basic_pme R
    | basic_pme_ext   : forall R S c, basic_cst R c -> basic_pme R -> S ~ pme (R cup sg c) -> basic_pme S.

  Proposition basic_pme_id R S : R ~ S -> basic_pme R -> basic_pme S.
  Proof.
    intros H1 H2; generalize H2 S H1; clear H1 S H2.
    induction 1 as [ R HR | R S c Hc HR H1 H2 ]; intros Q HQ.
    constructor 1; apply id_transitive with (2 := HR); auto.
    constructor 2 with (1 := Hc); auto.
    apply id_transitive with (2 := H2); auto.
  Qed.
  
  Proposition basic_pme_pme : forall R, basic_pme R -> is_pme R.
  Proof. 
    induction 1 as [ ? G | ? ? ? ? ? ? G ];
    apply id_symetric in G;
    apply is_pme_id with (1 := G); auto.
  Qed.
  
  Hint Resolve basic_pme_pme.
  
  Let gr_dimage_lt0 X (f : nat*X -> Prop)  : gr_dimage f (fun i => i < 0) ~ empty.
  Proof.
    split; auto.
    intros k ( x & Hx & _ ); contradict Hx; auto with arith.
  Qed.

  Definition basic_ext (f : _ -> Prop) n c := f (n,c) -> basic_cst (pme (gr_dimage f (fun i => i < n))) c.

  (* by induction, we extract a finite sequence that computes the basic pme *)
  
  Proposition basic_pme_seq R : basic_pme R -> 
            exists n, exists f, bounded_graph n f
                             /\ R ~ pme (gr_dimage f (fun i => i < n)) 
                             /\ forall k c, k < n -> basic_ext f k c.
  Proof.
    induction 1 as [ R HR | R Q c Hc HR IH HQ ].
    
    exists 0; exists empty.
    split; auto.
    split.
    apply id_transitive with (1 := HR).
    apply pme_id.
    apply id_symetric; auto.
    intros k ? Hk; contradict Hk; auto with arith.
    
    destruct IH as ( n & f & H1 & H2 & H3 ).
    
    exists (S n).
    exists (f cup sg (n,c)).
    split; auto.
    split.
    apply id_transitive with (1 := HQ).
    apply id_symetric in H2.
    apply id_symetric. 
    apply id_transitive with (1 := pme_id (gr_dimage_cup1 _ _ _ _ _)).
    apply cl_extend_id; auto; try (red; unfold cst; auto).
    apply id_transitive with (2 := is_pme_fixpoint (basic_pme_pme HR)).
    apply id_transitive with (2 := H2).
    apply pme_id.
    split.
    apply inc_transitive with (1 := gr_dimage_bounded _ _ _ _ H1); auto.
    apply gr_dimage_inc_src; intro; auto.
    apply pme_id, gr_dimage_sg_gr; auto.
    intros k d Hk Hd.
    apply bin_union_sg in Hd.
    assert (gr_dimage f (fun i => i < k) ~ gr_dimage (f cup sg (n,c)) (fun i => i < k) ) as H.
      split; auto.
      apply inc_transitive with (1 := proj1 (gr_dimage_cup1 _ _ _ _ _)).
      apply bin_union_inc_l; auto.
      intros c' [ x [ H4 H5 ] ].
      apply in_sg_eq in H5.
      injection H5; intros _ H6; contradict H6; omega.
    apply pme_id in H.
    apply basic_id with (1 := H).
    destruct Hd as [ Hd | Hd ].
    
    apply H3; auto.
    apply (proj1 H1) with (x := d); auto.
    
    apply in_sg_eq in Hd.
    injection Hd; clear Hd; intros Hn Hd.
    rewrite <- Hn, <- Hd.
    apply basic_id with (1 := H2).
    auto.
  Qed.
  
  Let pme_lemma (R S : constraint L -> Prop) : pme (R cup S) ~ pme (pme R cup S).
  Proof. split; auto. Qed.
  
  (* from any finite sequence with the basic_cst property, we obtain a basic pme *)
  
  Proposition seq_basic_pme n f : bounded_graph n f 
                               -> (forall k c, k < n -> basic_ext f k c) 
                               -> basic_pme (pme (gr_dimage f (fun i => i < n))).
  Proof.
    generalize f; clear f.
    induction n as [ | n IHn ]; intros f Hf H.
    constructor 1.
    apply pme_id; auto.
    set (g := restriction n f).
    assert (bounded_graph n g) as Hg.
      apply restriction_bounded' with (2 := Hf); auto.
    assert (basic_pme (pme (gr_dimage g (fun i => i < n)))) as Hg'.
      apply (IHn _ Hg).
      intros k c Hk [ _ Hc ].
      apply H in Hc; auto.
      apply basic_id with (2 := Hc), pme_id, restriction_dimage'.
      intros ? ?; apply lt_trans with (2 := Hk); auto.
    destruct (proj2 Hf _ (le_refl _)) as [ c Hc ].
    apply gr_dimage_sg' in Hc.
    constructor 2 with (c := c) (2 := Hg').
    apply basic_id with (2 := H n _ (le_refl _) Hc).
    apply pme_id, restriction_dimage'; auto.
    apply id_transitive with (2 := pme_lemma _ _).
    apply pme_id.
    clear IHn H Hg'.
    
    split.
    
    intros d ( m & Hm1 & Hm2 ).
    destruct (le_lt_dec n m).
    cutrewrite (m=n) in Hm2.
    rewrite (bounded_graph_functional _ _ _ Hf _ _ _ Hc Hm2); pred_solve.
    apply le_S_n in Hm1; auto with arith.
    apply bin_union_inc_r1; exists m; repeat (split; auto).
    
    pred_solve.
    apply inc_transitive with (1 := restriction_dimage_inc _ _ _ _).
    apply gr_dimage_inc_src.
    intro; auto with arith.
    exists n; auto.
  Qed.
  
  Definition simple_seq f : Prop := function_graph f /\ forall n c, basic_ext f n c.
  Definition simple_pme (R : _ -> Prop) : Prop := exists f, R ~ pme (gr_fst f) /\ simple_seq f.
  
  Proposition basic_inc_simple_pme : basic_pme inc simple_pme.
  Proof.
    intros R H.
    apply basic_pme_seq in H.
    destruct H as (n & f & H1 & H2 & H3).
    exists (extension _ n f (eps~~eps)).
    split.
    
    generalize (extension_gr_fst _ _ _ H1 (eps~~eps)); intros H4.
    apply pme_id in H4.
    apply id_transitive with (1 := H2),
          id_symetric,
          id_transitive with (1 := H4),
          id_transitive with (1 := pme_eps_1 _),
          pme_id,
          bounded_graph_fst; auto.
    
    split.
    apply extension_function; auto.
    
    intros k c Hkc.
    red in Hkc.
    destruct (le_lt_dec n k) as [ H | H ].
    rewrite -> Hkc. apply basic_eps; auto.
    assert (gr_dimage (extension _ n f (eps~~eps)) (fun i => i < k) ~ gr_dimage f (fun i => i < k)) as Hf.
      apply (extension_dimage2 _ _ _ H1 (eps~~eps) (fun i => i < k)).
      intros ? ?; apply lt_trans with (2 := H); auto.
    apply pme_id in Hf.
    apply id_symetric in Hf.
    apply basic_id with (1 := Hf).
    apply H3; auto.
  Qed.
  
  Proposition simple_is_pme : simple_pme inc (@is_pme _).
  Proof.
    intros R ( ? & H & _ ).
    apply id_symetric in H.
    apply is_pme_id with (1 := H); auto.
  Qed.
  
  Hint Immediate basic_inc_simple_pme simple_is_pme.
  
  Proposition basic_is_pme : basic_pme inc (@is_pme _).
  Proof. apply inc_transitive with (2 := simple_is_pme), basic_inc_simple_pme. Qed.
  
  Proposition simple_is_closed R : simple_pme R -> pme R inc R.
  Proof. intros; apply is_pme_postfixpoint, simple_is_pme; auto. Qed.
  
  Proposition basic_is_closed R : basic_pme R -> pme R inc R.
  Proof. intros; apply is_pme_postfixpoint, basic_is_pme; auto. Qed.

  Section Basic_sequence.

    Let pme_lem1 (A B C : constraint L -> Prop) c : pme A ~ pme B -> B cup (sg c) ~ C -> pme (A cup sg c) ~ pme C.
    Proof.
      intros [ H1 H2 ] [ H3 H4 ].
      apply pme_mono in H3.
      split.
      apply pme_min; auto.
      apply inc_transitive with (2 := H3).
      pred_solve.
      apply inc_transitive with (1 := pme_inc _).
      apply inc_transitive with (1 := H1); auto.
      apply pme_inc; pred_solve.
      apply pme_min; auto.
      apply inc_transitive with (1 := H4).
      pred_solve.
      apply inc_transitive with (1 := pme_inc _).
      apply inc_transitive with (1 := H2); auto.
      apply pme_inc; pred_solve.
    Qed.

    Variable C0 : constraint L -> Prop.
    Hypothesis H0 : basic_pme (pme C0).
    
    Variable C : nat -> constraint L -> Prop.
    Variable ce : graph nat (constraint L).
    
    Hypothesis Hc: function_graph ce.
    
    Hypothesis HC0 : C 0 ~ C0.
    Hypothesis HC1 : forall i c, ce (i,c) -> C (S i) ~ C i cup sg c.
    Hypothesis HC2 : forall i c, ce (i,c) -> basic_cst (pme (C i)) c.

    Let Hlt0 : (fun i => i < 0) ~ empty.
    Proof.
      split; auto.
      intros x Hx; contradict Hx; auto with arith.
    Qed.
    
    Theorem basic_seq_is_simple : simple_pme (pme (union C)).
    Proof.
      destruct (basic_pme_seq H0) as (n0 & f0 & Hf1 & Hf2 & Hf3).

      set (g p := match p with (n,c) => (n < n0  /\ f0 (n,c))
                                     \/ (n0 <= n /\ ce (n-n0,c)) 
                  end).
 
      assert ( function_graph g ) as Hg2.
        intros n.
        destruct (le_lt_dec n0 n) as [ H | H ].
        destruct (Hc (n-n0)) as [ d Hd ].
        apply gr_dimage_sg' in Hd.
        exists d.
        split.
        intros e ( m & He2 & [ [He3 He4] | [He3 He4] ] );
        apply in_sg_eq in He2.
        contradict He3; omega.
        rewrite <- He2 in He4.
        rewrite (function_graph_functional _ _ _ Hc _ _ _ Hd He4).
        auto.
        apply sg_inc.
        exists n.
        split; auto.
        right; auto.
        destruct (proj2 Hf1 _ H) as [ d Hd ].
        apply gr_dimage_sg' in Hd.
        exists d.
        split.
        intros e ( m & He2 & [ [He3 He4] | [He3 He4] ] );
        apply in_sg_eq in He2.
        rewrite <- He2 in He4.
        rewrite (bounded_graph_functional _ _ _ Hf1 _ _ _ Hd He4).
        auto.
        contradict He3; omega.
        apply sg_inc.
        exists n.
        split; auto.
        left; auto.

      assert (forall n, n < n0 -> gr_dimage g (fun i => i < n) ~ gr_dimage f0 (fun i => i < n)) as Hg0.
        intros n Hn.
        split.
        intros d ( z & Hz1 & [ [ Hz2 Hz3 ] | [ Hz2 Hz3 ] ] ).
        exists z; auto.
        contradict Hz2; omega.
        intros d ( z & Hz1 & Hz2 ).
        exists z.
        split; auto.
        left.
        split; auto; omega.

     assert (forall n, gr_dimage g (fun i => i < n+n0) ~ gr_dimage f0 (fun i => i < n0) cup gr_dimage ce (fun  i => i < n)) as Hg1.
        intros n; split.
        intros x ( y & Hy1 & [ ( Hy2 & Hy3 ) | ( Hy2 & Hy3 ) ] ).
        apply bin_union_inc_r1.
        exists y; auto.
        apply bin_union_inc_r2.
        exists (y-n0); split; auto; omega.
        pred_solve.
        intros x ( y & Hy1 & Hy2 ).
        exists y.
        split. omega.
        left; auto.
        intros x ( y & Hy1 & Hy2 ).
        exists (y+n0).
        split. omega.
        right. 
        split. omega.
        cutrewrite (y+n0-n0 = y); auto; omega.
 
      assert (forall n, pme (C n) ~ pme (gr_dimage g (fun i => i < n+n0))) as Hg3.
        induction n as [ | n Hn ].

        apply pme_id in HC0.
        apply id_transitive with (1 := HC0).
        apply id_transitive with (1 := Hf2).
        apply pme_id.
        apply id_symetric.
        apply id_transitive with (1 := Hg1 _).
        apply gr_dimage_id_src with (f := ce) in Hlt0.
        destruct Hlt0  as [ H1 H2 ]; split; pred_solve.
        apply inc_transitive with (1 := H1).
        generalize (proj1 (gr_dimage_empty_src _ _ ce)); intros H.
        apply inc_transitive with (1 := H); auto. 
        
        destruct (Hc n) as [ d Hd ].
        apply gr_dimage_sg' in Hd.
        generalize Hd; intros Hd'.
        apply HC1 in Hd'.
        apply pme_id in Hd'.
        apply id_transitive with (1 := Hd').
        apply pme_lem1 with (1 := Hn).
        apply id_symetric.
        cutrewrite (S n + n0 = S (n+n0)).
        apply gr_dimage_sect; auto.
        apply function_graph_functional; auto.
        right.
        split.
        omega.
        cutrewrite (n+n0-n0=n); auto.
        omega.
        omega.

      exists g.
      split.
      split; apply cl_inc; red; auto.
      intros d Hd.
      apply union_sg in Hd.
      destruct Hd as [ x Hx ].
      apply (pme_inc (C x)) in Hx.
      apply Hg3 in Hx.
      generalize d Hx; clear d Hx.
      apply pme_mono.
      apply gr_dimage_inc_src.
      intro; auto.
      intros d ( n & _ & Hn ).
      apply (pme_mono (union_inc_r _ _ C (S n))).
      apply Hg3.
      apply pme_inc.
      exists n.
      split; auto.
      omega.

      split; auto.
      intros n d [ [ Hd1 Hd2 ] | [ Hd1 Hd2 ] ].
      generalize (Hg0 _ Hd1); intros Hd3.
      apply pme_id in Hd3.
      apply id_symetric in Hd3.
      apply basic_id with (1 := Hd3).
      apply Hf3; auto.
      generalize (Hg3 (n-n0)); intros Hg.
      cutrewrite (n-n0+n0 = n) in Hg.
      apply basic_id with (1 := Hg).
      apply HC2; auto.
      omega.
    Qed.

  End Basic_sequence.

End Elem_simpl.
        
      
      


 