(**************************************************************)
(*   The Strong Completeness of Partial Monoidal Boolean BI   *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Import list_fall_exst.
Require Import predicate.
Require Import closure.
Require Import maps.
Require Import carrier. 
Require Import finite.

Require Import labels.
Require Import formula.
Require Import words.
Require Import substitutions.
Require Import constraints.
Require Import css.
Require Import pme.

Require Import branch.
Require Import tableaux.

Require Import bbi_expansion.
Require Import bbi_exp_props.
Require Import bbi_closed_branch.

Definition bbi_tab := tableau CSS_id bbi_expansion.

Definition bbi_has_closed_tableau := has_closed_tableau CSS_id bbi_expansion bbi_closed_branch.

Definition bbi_has_no_closed_tableau css := finite_CSS css /\ ~ has_closed_tableau CSS_id bbi_expansion bbi_closed_branch css.

Proposition bbi_tab_CSS css t : CSS css -> bbi_tab css t -> carrier t inc CSS.
Proof. intros Hcss ?; apply (tableau_prop_conj _ _ bbi_expansion _ _ Hcss CSS_id_id bbi_expansion_CSS); auto. Qed.
 
Proposition bbi_tab_finite_CSS css t : finite_CSS css -> bbi_tab css t -> carrier t inc finite_CSS.
Proof. intros Hcss ?; apply (tableau_prop_conj _ _ bbi_expansion _ _ Hcss fCSS_id bbi_expansion_finite_CSS); auto. Qed.

Proposition bbi_tab_increase css t : bbi_tab css t -> forall cssi, carrier t cssi -> CSS_leq css cssi.
Proof.
  apply tableau_prop_conj; auto.
  
  intros x y H0 H1.
  apply CSS_leq_trans with (2 := CSS_id_imp_leq _ _ H0); auto.
  
  intros b lb H0 H1 x Hx.
  apply CSS_leq_trans with (1 := H0).  
  apply bbi_expansion_increase' with (1 := H1); auto.
Qed.

Proposition bbi_tab_CSS_id css1 css2 t : CSS_id css1 css2 -> bbi_tab css1 t -> bbi_tab css2 t.
Proof.
  intro; apply tableau_eq; auto.
  apply CSS_id_trans.
Qed.

Proposition bbi_tab_comp b0 l1 l2 b t : bbi_tab b t 
                                     -> bbi_tab b0 (l1 ++ b :: l2)
                                     -> bbi_tab b0 (l1 ++ t ++ l2).
Proof. 
  apply tableau_comp; auto.
  apply CSS_id_trans.
Qed.

Section BBI_Tableaux_basic.

  Let bbi_tableau_exp css lcss : bbi_expansion css lcss -> bbi_tab css lcss.
  Proof.
    intros H.
    cutrewrite (lcss = nil++lcss++nil).
    apply tableau_next with (2 := H); auto.
    apply tableau_start; auto.
    rewrite <- app_nil_end; auto.
  Qed.

  Variable X : stm -> Prop.
  Variable C : cst -> Prop.

  Hint Resolve bbi_exp_Tnot bbi_exp_Fnot bbi_exp_Tand bbi_exp_Fand.
  Hint Resolve bbi_exp_TI  bbi_exp_Tmult  bbi_exp_Fmult  bbi_exp_Twand  bbi_exp_Fwand.

  Ltac solve := intros; apply bbi_tableau_exp; eauto.
  
  Proposition bbi_tableau_exp_Tnot A m : X (T !A # m) -> bbi_tab (X,C) ((X cup sg (F A # m), C cup empty)::nil).
  Proof. solve. Qed. 

  Proposition bbi_tableau_exp_Fnot A m : X (F !A # m) -> bbi_tab (X,C) ((X cup sg (T A # m), C cup empty)::nil).
  Proof. solve. Qed.
 
  Proposition bbi_tableau_exp_Tand A B m : X (T A && B # m) -> bbi_tab (X,C) ((X cup (sg (T A # m) cup sg (T B # m)), C cup empty)::nil).
  Proof. solve. Qed.
  
  Proposition bbi_tableau_exp_Fand A B m : X (F A && B # m) -> bbi_tab (X,C) ((X cup sg (F A # m), C cup empty)::
                                                                              (X cup sg (F B # m), C cup empty)::nil).
  Proof. solve. Qed.

  Proposition bbi_tableau_exp_TI m : X (T I # m) -> bbi_tab (X,C) ((X cup empty, C cup sg (eps ~~ m))::nil).
  Proof. solve. Qed.

  Proposition bbi_tableau_exp_Tmult A B a b m : X (T A * B # m) -> a<> b -> ~ alphabet C a -> ~ alphabet C b 
                                             -> bbi_tab (X,C) ((X cup (sg (T A # l a) cup sg (T B # l b)), C cup sg (l a o l b ~~ m))::nil).
  Proof. solve. Qed.

  Proposition bbi_tableau_exp_Fmult A B x y m : X (F A * B # m) -> pme C (x o y ~~ m)
                                             -> bbi_tab (X,C) ((X cup sg (F A # x), C cup empty)::
                                                               (X cup sg (F B # y), C cup empty)::nil).
  Proof. solve. Qed.

  Proposition bbi_tableau_exp_Twand A B x y m : X (T A -* B # m) -> pme C (x o m ~~ y)
                                             -> bbi_tab (X,C) ((X cup sg (F A # x), C cup empty)::
                                                               (X cup sg (T B # y), C cup empty)::nil).
  Proof. solve. Qed.

  Proposition bbi_tableau_exp_Fwand A B a b m : X (F A -* B # m) -> a<> b -> ~ alphabet C a -> ~ alphabet C b 
                                             -> bbi_tab (X,C) ((X cup (sg (T A # l a) cup sg (F B # l b)), C cup sg (l a o m ~~ l b))::nil).
  Proof. solve. Qed.

End BBI_Tableaux_basic.

Section BBI_Tableaux_props_1.

  Proposition bbi_tab_CSS_inc css1 css2 t : CSS_inc css1 css2 
                                         -> bbi_tab css1 t 
                                         -> exists t', bbi_tab css2 t' 
                                                    /\ listbranch_eq CSS_inc t t'.
  Proof.
    intros G1 G2.
    apply tableau_prop_inc with (4 := G2); auto.

    intros (X1,C1) (X2,C2) (X3,C3) (X4,C4) [ H1 H2 ] [ H3 H4 ] (H5 & H6 & H7).
    split.
    apply id_transitive with (2 := H3), id_transitive with (2 := H5), id_symetric; auto.
    split.
    apply pme_id in H2.
    apply pme_id in H4.
    apply inc_transitive with (2 := proj1 H4),
          inc_transitive with (2 := H6), H2.
    apply alphabet_id in H2.
    apply alphabet_id in H4.
    apply id_transitive with (2 := H4), id_transitive with (2 := H7), id_symetric; auto.

    intros (X,C) (X',C') lb  (H1 & H2 & H3) [ r p Hr Hp ].
    exists (bbi_expand (X',C') (bbi_inst r p)).
    split; auto.
    apply bbi_fireable_inc with (1 := H1) (2 := H2) (4 := Hr) in Hp; auto.
    constructor 1 with (2 := Hp); auto.
    apply listbranch_eq_expand_CSS_inc; auto.
    repeat (split; auto).
  Qed.

End BBI_Tableaux_props_1.

Section BBI_Tableaux_props_2.

  Variable sigma : label -> label.
  Hypothesis sigma_inj : injective sigma.
    
  Let s a := l (sigma a).

  Proposition bbi_tab_subst css t : bbi_tab css t -> bbi_tab (subst_CSS s css) (map (subst_CSS s) t).
  Proof.
    generalize CSS_id_trans; intros ?.
    apply tableau_prop_fun; auto.
    apply bbi_expansion_subst_inj; auto.
  Qed.

End BBI_Tableaux_props_2.

Section BBI_Tableaux_props_3.

  Variables Xe : stm -> Prop.
  Variables Ce : cst -> Prop.

  Ltac it1 x := apply inc_transitive with (1 := x).
  Ltac it2 x := apply inc_transitive with (2 := x).

  Ltac reduce := repeat
    match goal with 
        [ |- bbi_tab_constraints ( _ ++ _ ) inc _ ] => apply inc_transitive with (1 := proj1 (bbi_tab_constraints_app _ _))
      | [ |- _ inc bbi_tab_constraints ( _ ++ _ ) ] => apply inc_transitive with (2 := proj2 (bbi_tab_constraints_app _ _))
      | [ |- bbi_tab_constraints ( _ :: _ ) inc _ ] => simpl
      | [ |- _ inc bbi_tab_constraints ( _ :: _ ) ] => simpl
      | [ |- ?x cup _ inc ?x cup _ ] => apply bin_union_monotone; auto
      | [ |- ?x cap _ inc ?x cap _ ] => apply bin_inter_monotone; auto
      | [ |- _ cup ?x inc _ cup ?x ] => apply bin_union_monotone; auto
      | [ |- _ cap ?x inc _ cap ?x ] => apply bin_inter_monotone; auto
    end.

  Proposition bbi_tab_cup : forall X0 C0 t, bbi_tab (X0,C0) t 
                                         -> alphabet (bbi_tab_constraints t) cap alphabet Ce inc alphabet C0
                                         -> bbi_tab (Xe cup X0, Ce cup C0) (bbi_expand (Xe,Ce) t).
  Proof.
    generalize CSS_id_trans; intros Htrans.
    intros X0 C0;
    induction 1 as [ (X1,C1) [ Hb1 Hb2 ] | l1 (X1,C1) l2 lb l' H1 H2 H3 H4 ]; intros Halpha.
    
    simpl bbi_expand; unfold bbi_expand_1.
    apply tableau_eq with (b0 := (Xe cup X0, Ce cup C0))
                          (t0 := (Xe cup X0, Ce cup C0)::nil); auto.
    simpl; split; auto; simpl; split; auto.
    apply tableau_start; auto.
    
    match goal with [ _: ?x -> bbi_tab _ _ |- _ ] => assert x as H5 end.
      it2 Halpha. 
      reduce. 
      apply bbi_tab_constraints_id, alphabet_id in H2.
      it2 (proj1 H2).
      apply alphabet_inc.
      reduce.
      apply bbi_expansion_increase with (1 := H1).   
    generalize (H4 H5); clear H4; intros H4.

    apply tableau_eq with (b0 := (Xe cup X0, Ce cup C0))
                          (t0 := (bbi_expand (Xe,Ce) l1)++(bbi_expand (Xe,Ce) lb)++(bbi_expand (Xe,Ce) l2)); auto.
    apply bbi_expand_id with (css1 := (Xe,Ce)) (css2 := (Xe,Ce)) in H2; auto.
    apply listbranch_eq_tran with (3 := H2); auto.
    do 2 rewrite bbi_expand_app; auto.

    rewrite bbi_expand_app in H4; simpl in H4.
    destruct bbi_expansion_cup with (Xe := Xe) (Ce := Ce) (1 := H1) as ( lb' & H6 & H7 ).

    destruct (bbi_tab_increase _ _ H3 (X1,C1)) as [ _ HC1 ].
    apply carrier_app; pred_solve.
    simpl in HC1; apply alphabet_inc in HC1.
    it2 HC1.
    it2 Halpha.
    reduce.
    apply alphabet_inc; auto.
    apply bbi_tab_constraints_id in H2.
    it2 (proj1 H2).
    reduce.
    apply inc_transitive with (2 := bin_union_inc_r2 _ _ _).
    reduce; auto.
    
    unfold bbi_expand_1 in H4.
    apply tableau_next with (2 := H6) in H4; auto.
    apply tableau_eq with (6 := H4); auto.
  Qed.

End BBI_Tableaux_props_3.

(* Here is the proof of Theorem which is developped in the Appendix of the paper *)

Section BBI_closed_Tableaux_mono.

  Variables X0 X1 : stm -> Prop.
  Variables C0 C1 : cst -> Prop.

  Hypothesis H0 : finite_CSS (X0,C0).
  Hypothesis H1 : finite_CSS (X1,C1).
  Hypothesis H21 : X0 inc X1.
  Hypothesis H22 : C0 inc C1.

  Proposition bbi_has_closed_tableau_monotonic : bbi_has_closed_tableau (X0,C0) -> bbi_has_closed_tableau (X1,C1).
  Proof.
    generalize CSS_id_trans; intros Htrans.
    destruct H0 as ( H01 & H02 & H03 ).
    destruct H1 as ( H11 & H12 & H13 ).
    destruct (label_hole_injection (alphabet C0) (alphabet C1)) as (sigma & Hs1 & Hs2); auto.
    assert (forall x, alphabet C0 x -> (fun a => l (sigma a)) x = (fun a => l a) x) as Hs3.
      intros x Hx. f_equal; destruct (Hs2 x); tauto.
    assert (forall m, carrier m inc alphabet C0 -> subst_w (fun a => l (sigma a)) m = m) as Hs4.
      intros m Hm.
      rewrite -> (subst_carrier_eq _ _ Hs3 Hm). 
      apply subst_w_id.
    assert (forall x, alphabet C1 (sigma x) -> alphabet C0 (sigma x)) as Hs5.
      intros x HX.
      destruct (Hs2 x) as [ [ G1 G2 ] | [ G1 G2 ] ].
      rewrite G2; auto.
      contradict G2; auto.

    (* we have a substitution of letters that 
          - is injective
          - will rename any letter in C1\C0 into some letter outside C1 
          - is stable on the letters of C0
     *)

    intros [ t [ Ht1 Ht2 ] ].
    generalize Ht1; intros Ht3.
    apply bbi_tab_subst with (1 := Hs1) in Ht3.
    assert (CSS_id (subst_CSS (fun a => l (sigma a)) (X0,C0)) (X0,C0)) as Ht5. 
      apply CSS_id_trans with (2 := subst_CSS_id' _).
      apply subst_CSS_alpha_id; auto.
    set (st := map (subst_CSS (fun a => l sigma a)) t); fold st in Ht3.
    apply tableau_eq with (b1 := (X0,C0)) (t1 := st) in Ht3; auto.
    assert (alphabet (bbi_tab_constraints st) cap alphabet C1 inc alphabet C0) as Ht7.
      unfold st.
      apply bbi_tab_constraints_subst.
      intros x y Hy; apply bin_inter_sg in Hy.
      destruct Hy as [ [ Hy1 | [] ] Hy2 ].
      rewrite <- Hy1 in Hy2 |- *; auto.
    apply bbi_tab_cup with (Xe := X1) (Ce := C1) in Ht3; auto.
    exists (bbi_expand (X1,C1) st).
    split.
    apply tableau_eq with (6 := Ht3); auto.
    apply bbi_cb_expand_carrier; auto.
    apply bbi_cb_subst_carrier; auto.
  Qed.

End BBI_closed_Tableaux_mono.

Section BBI_closed_tableau_pme_monotonic.

  Variables X0 X1 : stm -> Prop.
  Variables C0 C1 : cst -> Prop.

  Hypothesis H0 : finite_CSS (X0,C0).
  Hypothesis H1 : finite_CSS (X1,C1).
  Hypothesis H21 : X0 inc X1.
  Hypothesis H22 : C0 inc pme C1.

  Proposition bbi_has_closed_tableau_pme_monotonic : bbi_has_closed_tableau (X0,C0) -> bbi_has_closed_tableau (X1,C1).
  Proof.
    intros H.
    apply bbi_has_closed_tableau_monotonic with (X1 := X1) (C1 := C0 cup C1) in H; auto.
    destruct H as ( t & Ht1 & Ht2 ).
    apply bbi_tab_CSS_inc with (css2 := (X1,C1)) in Ht1.
    destruct Ht1 as ( t' & Ht3 & Ht4 ).
    exists t'. 
    split; auto.
    apply bbi_cb_inc_carrier with (2 := Ht2); auto.
    split; auto.
    split; auto.
    split; auto.
    apply inc_transitive with (1 := proj1 (alphabet_pme _)),
          inc_transitive with (2 := proj2 (alphabet_pme _)),
          alphabet_inc; auto.
    destruct H0 as [ ? [] ].
    destruct H1 as [ ? [] ].
    auto.
  Qed.

End BBI_closed_tableau_pme_monotonic.
 
Corollary bbi_has_closed_tableau_eps X C : finite_CSS (X,C) -> bbi_has_closed_tableau (X,C cup sg (eps~~eps)) -> bbi_has_closed_tableau (X,C).
Proof.
  intros H; generalize H; intros [ ? [ ? ? ] ].
  apply bbi_has_closed_tableau_pme_monotonic; auto.
Qed.

Section bbi_proof_subst.

  Variable so : label -> label.
  Variable X : stm -> Prop.
  Variable C : cst -> Prop.

  Hypothesis HC : forall x, alphabet C x \/ ~ alphabet C x.  
  Hypothesis HXC : CSS (X,C).

  Hypothesis Hso1 : forall a b, ~ alphabet C a -> ~ alphabet C b -> so a = so b -> a = b.
  Hypothesis Hso2 : pr_image so (pd_image so (alphabet C)) inc alphabet C.

  Let Hso1' D : C inc D -> forall a b, ~ alphabet D a -> ~ alphabet D b -> so a = so b -> a = b.
  Proof.
    intros H ? ? Ha Hb; apply Hso1; [ contradict Ha | contradict Hb ]; apply alphabet_inc with (1 := H); auto.
  Qed.

  Let Hso2' D : C inc D -> pr_image so (pd_image so (alphabet D)) inc alphabet D.
  Proof.
    intros H x (y & ( z & H1 & H2) & H3).
    destruct (HC z) as [ | H4 ].
    apply alphabet_inc with (1 := H).
    apply Hso2.
    exists y; split; auto.
    exists z; split; auto.
    destruct (HC x) as [ | H5 ].
    apply alphabet_inc with (1 := H); auto.
    rewrite (Hso1 _ _ H5 H4); auto.
    rewrite H2; auto.
  Qed.

  Let s x := l (so x).
  
  Proposition bbi_proof_subst t :  bbi_tab (X, C) t ->  bbi_tab (subst_CSS s (X,C)) (map (subst_CSS s) t).
  Proof.
    generalize CSS_id_trans; intros Htrans.
    induction 1 as [ (X1,C1) [ Hb1 Hb2 ] | l1 (X1,C1) l2 lb l' H1 H2 H3 H4 ].
    
    simpl.
    constructor 1.
    split.
    unfold subst_pstm; auto.
    unfold subst_dimage; auto.
    
    rewrite map_app in H4; simpl in H4.
    apply listbranch_eq_map with (f := subst_CSS s) in H2; auto.
    destruct (bbi_tab_increase _ _ H3 (X1,C1)) as [ _ H5 ].
    apply carrier_app; pred_solve.
    destruct (bbi_expansion_subst_stable) with (sigma := so) (3 := H1) as ( lb' & H7 & H8 ); simpl;
      try (apply alphabet_inc with (1 := H5); auto).
    apply Hso1'; auto.
    apply Hso2'; auto.    
    apply tableau_eq with (b0 := subst_CSS s (X,C))
                          (t0 := map (subst_CSS s) l1 ++ lb' ++ map (subst_CSS s) l2); auto.
    apply listbranch_eq_tran with (3 := H2); auto.
    do 2 rewrite map_app; auto.
    apply tableau_next with (3 := H4); auto.
  Qed.

End bbi_proof_subst.

Section BBI_proofs.

  Variable a b : label.
  Variable A : formula.

  Proposition bbi_has_proof_ab : bbi_has_closed_tableau (sg (F A # l a), sg (l a ~~ l a)) -> bbi_has_closed_tableau (sg (F A # l a), sg (l a ~~ l b)).
  Proof.
    apply bbi_has_closed_tableau_pme_monotonic; auto; 
    [  split; auto; apply CSS_sg_l | pred_solve ]; 
    apply pme_lft with (y := l b), pme_inc; auto.
  Qed.

  Proposition bbi_has_proof_aa : bbi_has_closed_tableau (sg (F A # l a), sg (l a ~~ l b)) -> bbi_has_closed_tableau (sg (F A # l a), sg (l a ~~ l a)).
  Proof.
    generalize CSS_id_trans; intros Htrans.
    intros ( t & Ht1 & Ht2 ).
    apply bbi_proof_subst with (so := subst_one_letter b a) in Ht1.
    simpl in Ht1.
    exists (map (subst_CSS (fun x => l (subst_one_letter b a x))) t).
    split.
    apply tableau_eq with (6 := Ht1); auto.
    split.

    unfold subst_pstm.
    apply id_transitive with (1 := pd_image_sg _ _ _ _).
    simpl; rewrite subst_one_letter_prop_2; auto.

    unfold subst_dimage.
    apply id_transitive with (1 := pd_image_sg _ _ _ _).
    simpl; rewrite subst_one_letter_prop_0, subst_one_letter_prop_2; auto.

    apply bbi_cb_subst_carrier; auto.

    apply finite_in_dec; auto.

    intros x y Hx Hy.
    rewrite subst_one_letter_prop_1, subst_one_letter_prop_1; auto.
    contradict Hy; rewrite <- Hy; apply alphabet_sg; pred_solve.
    contradict Hx; rewrite <- Hx; apply alphabet_sg; pred_solve.

    intros x (y & (z & H1 & H2) & H3).
    apply alphabet_sg, bin_union_sg in H1.
    apply alphabet_sg.
    destruct (label_eq_dec b x) as [ Hx | ].
    rewrite <- Hx; pred_solve.
    rewrite subst_one_letter_prop_1 in H3; auto.
    rewrite H3; clear dependent x.
    rewrite <- H2; clear H2.
    destruct H1 as [ [ H1 | [] ] | [ H1 | [] ] ].
    rewrite <- H1, subst_one_letter_prop_2; pred_solve.
    rewrite <- H1, subst_one_letter_prop_0; pred_solve.
  Qed.

End BBI_proofs.
 


