(**************************************************************)
(*   The Strong Completeness of Partial Monoidal Boolean BI   *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.
Require Import Permutation.

Require Import predicate.
Require Import closure.
Require Import carrier.

Require Import labels.
Require Import words.
Require Import constraints.
Require Import pme.

Require Import formula.
Require Import css.

Reserved Notation " m '|=' f " (at level 50, no associativity). 

Definition kripke_monotonic {X Y} (R : X -> X -> Prop) (f : X -> Y -> Prop) := forall x m n, R m n -> f m x -> f n x.

Section Model.

  Set Implicit Arguments.

  Variable K : Type.
  Variable meqK : constraint K -> Prop.
  Variable ivar : word K -> Var -> Prop.

  Notation "x 'meq' y" := (meqK (x~~y)) (at level 40, no associativity).

  Fixpoint forces (m : word _) f { struct f } := match f with 
    | % X    => ivar m X
    | I      => eps meq m
    | ! f    => m |= f -> False
    | f && g => m |= f /\ m |= g 
    | f *  g => exists x, exists y, x o y meq m /\ x |= f /\ y |= g
    | f -* g => forall x y, x o m meq y -> x |= f -> y |= g
  end where " m '|=' f " := (forces m f).

  Definition mdl_validity f := forall m, m meq m -> m |= f.

  Definition bbi_frame := is_pme meqK /\ kripke_monotonic (fun x y => x meq y) ivar.

  Hypothesis Hframe : bbi_frame. 
  
  Proposition mdl_equiv : meqK ~ pme meqK.
  Proof. destruct Hframe; split; auto. Qed.
  
  Proposition forces_monotonic : kripke_monotonic (fun x y => x meq y) forces.
  Proof.
    generalize Hframe; intros [ Hframe1 Hframe2 ].     
    intros f; induction f as [ v | [] | [] f Hf | [] f1 Hf1 f2 Hf2 ];
    intros m n H1 H2; simpl in H2 |- *. 
    
    apply Hframe2 with (1 := H1); auto.
    
    apply mdl_equiv in H1.
    apply mdl_equiv.
    apply pme_tran with (2 := H1), mdl_equiv; auto.
    
    contradict H2; apply Hf with (2 := H2); auto.
    apply mdl_equiv.
    apply mdl_equiv in H1; auto.
    
    destruct H2 as (x & y & H2 & H3 & H4).
    exists x; exists y; split; auto.
    apply mdl_equiv in H2.
    apply mdl_equiv, pme_tran with (1 := H2), mdl_equiv; auto.
    
    intros x y H3.
    apply mdl_equiv in H3.
    apply mdl_equiv in H1.
    apply H2, mdl_equiv, pme_el with (2 := H3); auto.
    
    destruct H2; split.
    apply Hf1 with (1 := H1); auto.
    apply Hf2 with (1 := H1); auto.
  Qed.

  Variable rho : label -> word K.

  Definition mdl_cst c := match c with (x,y) => subst_w rho x meq subst_w rho y end.

  Definition mdl_stm c := match c with T f # m =>   subst_w rho m |= f 
                                     | F f # m => ~ subst_w rho m |= f
                          end.
  
  Definition mdl_CSS K := match K with (X,C) => CSS (X,C) /\ X inc mdl_stm /\ C inc mdl_cst end.

  Proposition mdl_cst_pme : is_pme mdl_cst.
  Proof.
    apply meq_is_pme. 

    apply mdl_equiv, pme_eps; simpl; auto.

    simpl; intros ? ?; do 2 rewrite subst_o.
    intros H; apply mdl_equiv in H.
    apply mdl_equiv, pme_comm; auto.
    
    simpl; intros ? ?; rewrite subst_o.
    intros H; apply mdl_equiv, pme_sub in H.
    apply mdl_equiv; auto.

 
    intros ? ? ?; simpl; intros H1 H2.
    rewrite subst_o in H1.
    do 2 rewrite subst_o.
    apply mdl_equiv in H1.
    apply mdl_equiv in H2.
    apply mdl_equiv, pme_comp; auto.
    
    intros ? ? ? H1 H2.
    apply mdl_equiv in H1.
    apply mdl_equiv in H2.
    apply mdl_equiv, pme_tran with (1 := H1); auto.

    intros ? ? H.
    apply mdl_equiv in H.
    apply mdl_equiv, pme_sym; auto.
  Qed.

  Proposition meq_mdl_cst C : C inc mdl_cst -> pme C inc mdl_cst.
  Proof. apply pme_min, mdl_cst_pme. Qed.

  Proposition mdl_cst_meq X C m n : mdl_CSS (X,C) -> pme C (m~~n) -> subst_w rho m meq subst_w rho n.
  Proof.
    intros (HCSS,(HX,HC)) Hf.
    apply meq_mdl_cst in HC.
    apply HC in Hf; auto.
  Qed.

  Proposition meq_mdl_stm X C b f m : mdl_CSS (X,C) -> X (b,(f,m)) -> subst_w rho m meq subst_w rho m.
  Proof.
    intros H1 H4.
    apply mdl_cst_meq with (1 := H1).
    generalize H4; apply H1.
  Qed.

End Model.

Section Model_eq.

  Set Implicit Arguments.

  Variable K : Type.
  Variable meqK : constraint K -> Prop.
  Variable ivar : word K -> Var -> Prop.

  Notation " m '|=' f " := (forces meqK ivar m f).
   
  Let meq_mdl_eq_1 X C (rho rho' : label -> word K) : (forall a, alphabet C a -> rho a = rho' a) -> mdl_CSS meqK ivar rho' (X,C) -> mdl_CSS meqK ivar rho (X,C).
  Proof.
    intros Hr (H1 & H2 & H3).
    split; auto. split.

    intros (s,(f,k)) H; simpl. 
    rewrite subst_carrier_eq with (1 := Hr).
    apply H2 in H; apply H.
    apply H1 in H.
    apply pme_inc_alphabet; auto.
    
    intros (x,y) Hxy; simpl.
    rewrite subst_carrier_eq with (1 := Hr).
    rewrite subst_carrier_eq with (1 := Hr).
    apply H3 in Hxy; apply Hxy.
    apply pme_inc_alphabet, pme_mono with (R := sg (x~~y)); pred_solve; apply pme_rt with (x := x); auto.
    apply pme_inc_alphabet, pme_mono with (R := sg (x~~y)); pred_solve; apply pme_lft with (y := y); auto.
  Qed.
  
  Proposition meq_mdl_eq X C (rho rho' : label -> word K) : (forall a, alphabet C a -> rho a = rho' a) -> (mdl_CSS meqK ivar rho' (X,C) <-> mdl_CSS meqK ivar rho (X,C)).
  Proof.
    intros Hr; split; apply meq_mdl_eq_1; auto.
    intros; rewrite Hr; auto.
  Qed.

End Model_eq.

