(**************************************************************)
(*   The Strong Completeness of Partial Monoidal Boolean BI   *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import NPeano.
Require Import Max.
Require Omega.

Require Import replace.

Require Import predicate.
Require Import maps.
Require Import natseq.

Let nat_zero : forall n, { 0 = n } + { 0 < n }.
Proof. induction n; [ left | right ]; auto with arith. Qed.

Let leSn_disj := le_lt_or_eq.

Let ltSn_disj m n : m < S n -> m < n \/ m = n.
Proof. intros; apply le_lt_or_eq, lt_n_Sm_le; trivial. Qed.

Section Maxf.

  Variable f : nat -> nat.

  Fixpoint fnat_max n := match n with
    | 0 => 0
    | S n => max (f n) (fnat_max n)
  end.

  Theorem fnat_max_max n : forall i, i < n -> f i <= fnat_max n.
  Proof.
    induction n; intros i Hi; try (contradict Hi; auto with arith; fail).
    apply ltSn_disj in Hi.
    destruct Hi as [ Hi | Hi ].
    apply IHn in Hi.
    apply le_trans with (1 := Hi), le_max_r.
    rewrite Hi.
    apply le_max_l.
  Qed.

  Theorem fnat_max_exists n : 0 < n -> { i | i < n /\ f i = fnat_max n }.
  Proof.
    induction n; intros H; [ contradict H; auto with arith | clear H ].
    destruct n. exists 0. simpl. rewrite max_0_r. auto with arith.
    destruct IHn as ( i & IH1 & IH2 ). 
    auto with arith.
    simpl in IH2 |- *. rewrite <- IH2. 
    destruct (max_dec (f (S n)) (f i)) as [ H | H ]; 
    rewrite H; [ exists (S n) | exists i ]; auto with arith.
  Qed.

End Maxf.

Let img (f : nat -> nat) n k := exists i, i < n /\ f i = k.
Let sect n k := k < n.

Section Permut.

  Variable f : nat -> nat.
  Variable a b : nat.

  Definition permut n := match (eq_nat_dec a n,eq_nat_dec b n) with
    | (left _, _) => f b
    | (_, left _) => f a
    | _ => f n
  end.

  Proposition permut_a : permut a = f b.
  Proof.
    unfold permut.
    destruct (eq_nat_dec a a) as [ | H ]; trivial.
    exfalso. apply H; trivial.
  Qed.

  Proposition permut_b : permut b = f a.
  Proof.
    unfold permut.
    destruct (eq_nat_dec a b) as [ H | H ].
    rewrite H; trivial.
    destruct (eq_nat_dec b b) as [ | H' ]; trivial.
    exfalso. apply H'; trivial.
  Qed.

  Proposition permut_not_ab x : a <> x -> b <> x -> permut x = f x.
  Proof.
    intros Ha Hb. unfold permut.
    destruct (eq_nat_dec a x) as [ H | H ].
    contradiction (Ha H). clear H.
    destruct (eq_nat_dec b x) as [ H | H ].
    contradiction (Hb H). trivial.
  Qed.
    
  Hypothesis a_not_b : a <> b.

  Theorem permut_inj : fnat_inj f -> fnat_inj permut.
  Proof.
    intros H x y Hxy.
    unfold permut in Hxy.
    destruct (eq_nat_dec a x) as [ Hax | Hax ]; destruct (eq_nat_dec a y) as [ Hay | Hay ];
    destruct (eq_nat_dec b x) as [ Hbx | Hbx ]; destruct (eq_nat_dec b y) as [ Hby | Hby ];
    try (replace_with Hxy; auto with arith; fail); 
    try (rewrite <- Hbx; auto; fail);
    try (rewrite <- Hax; auto; fail); apply H in Hxy;
    rewrite Hxy in * |- *; auto with arith;
    try (rewrite <- Hax; auto; fail);
    try (rewrite <- Hbx; auto; fail);
    try (contradict Hby; auto);
    try (contradict Hbx; auto).
  Qed.
 
  Theorem permut_max n : a < n -> b < n -> fnat_max f n = fnat_max permut n.
  Proof.
    intros Ha Hb.
    destruct (nat_zero n) as [H | H].
    rewrite <- H. simpl. trivial.
    apply le_antisym.

    destruct (fnat_max_exists f _ H) as [ i [ H1 H2 ] ].
    rewrite <- H2. clear H H2.
    destruct (eq_nat_dec a i) as [ Hax | Hax ].
    rewrite <- Hax.
    apply le_trans with (2 := fnat_max_max _ _  _ Hb).
    rewrite permut_b. trivial.
    destruct (eq_nat_dec b i) as [ Hbx | Hbx ].
    rewrite <- Hbx.
    apply le_trans with (2 := fnat_max_max _ _  _ Ha).
    rewrite permut_a. trivial.
    apply le_trans with (2 := fnat_max_max _ _  _ H1).
    rewrite permut_not_ab; trivial.
    
    destruct (fnat_max_exists permut _ H) as [ i [ H1 H2 ] ].
    rewrite <- H2. clear H H2.
    destruct (eq_nat_dec a i) as [ Hax | Hax ].
    rewrite <- Hax.
    apply le_trans with (2 := fnat_max_max _ _  _ Hb).
    rewrite permut_a. trivial.
    destruct (eq_nat_dec b i) as [ Hbx | Hbx ].
    rewrite <- Hbx.
    apply le_trans with (2 := fnat_max_max _ _  _ Ha).
    rewrite permut_b. trivial.
    apply le_trans with (2 := fnat_max_max _ _  _ H1).
    rewrite permut_not_ab; trivial.
  Qed.
 
  Theorem permut_img n : a < n -> b < n -> img f n ~ img permut n.
  Proof.
    intros Ha Hb; split; intros x[ i [ H1 H2 ] ];
    destruct (eq_nat_dec a i) as [ Hai | Hai ].
    exists b; rewrite permut_b, Hai; auto.
    destruct (eq_nat_dec b i) as [ Hbi | Hbi ].
    exists a; rewrite permut_a, Hbi; auto.
    exists i; rewrite permut_not_ab; auto.
    exists b; rewrite <- Hai, permut_a in H2; auto.
    destruct (eq_nat_dec b i) as [ Hbi | Hbi ].
    exists a; rewrite <- Hbi, permut_b in H2; auto.
    exists i; rewrite permut_not_ab in H2; auto.
  Qed.

End Permut.

Section Maxf_img.

  Let img_sect_eq f n := sect n ~ img f n.

  Proposition imgsect_maxfn f n : img_sect_eq f n -> fnat_max f n = n-1.
  Proof.
    intros Hf.
    destruct (nat_zero n) as [ Hn | Hn ].
    rewrite <- Hn; simpl; trivial.
    apply le_antisym.
    destruct (fnat_max_exists f _ Hn) as [ a [ H1 H2 ] ].
    rewrite <- H2.
    assert (f a < n) as H.
    apply (proj2 Hf (f a)); exists a; auto. 
    rewrite <- pred_of_minus.
    apply le_pred in H.
    rewrite <- pred_Sn in H; auto.
    destruct (proj1 Hf (n-1)) as [ i [ H1 H2 ] ].
    red; auto with arith.
    rewrite <- H2.
    apply fnat_max_max; auto.
  Qed.

  Let maxf_step n : forall f, fnat_inj f -> 0 < n -> fnat_max f n < n -> f n = n -> img_sect_eq f n -> img_sect_eq f (S n).
  Proof.
    intros f Hf Hn Hm Hfn Hs.
    red in Hs.
    assert (forall i, i < n -> f i < f n) as Hnf'.
    intros i Hi.
    generalize (fnat_max_max f _ _ Hi). intros Hi'. 
    rewrite Hfn.
    apply le_lt_trans with (2 := Hm); auto.
    split.
    intros a Ha; red in Ha.
    apply lt_n_Sm_le in Ha.
    apply le_lt_or_eq in Ha.
    destruct Ha as [ Ha | Ha ].
    apply Hs in Ha.
    destruct Ha as [ i [ Hi1 Hi2 ] ]; exists i; auto with arith.
    exists n; rewrite Hfn; auto with arith.
    intros i ( j & Hi1 & Hi2 ).
    rewrite <- Hi2. red.
    apply lt_n_Sm_le in Hi1.
    apply le_lt_or_eq in Hi1.
    destruct Hi1 as [ Hi1 | Hi1 ].
    apply Hnf' in Hi1. rewrite <- Hfn; auto with arith.
    rewrite Hi1, Hfn; auto.
  Qed.

  Let maxf_rec n : forall f, fnat_inj f -> fnat_max f n < n -> img_sect_eq f n.
  Proof.
    induction n; intros f Hf Hn.

    simpl in Hn.
    contradict Hn; auto with arith.

    destruct (fnat_max_exists f (S n)) as [ a [ H1 H2 ] ]; try (auto with arith).
    assert (forall i, i < S n -> i <> a -> f i < f a) as Hf'.
      intros i Hi Ha.
      generalize (fnat_max_max f _ _ Hi); intros Hi'.
      apply le_lt_or_eq in Hi'.
      destruct Hi' as [ Hi' | Hi' ].
      rewrite H2; auto.
      rewrite <- H2 in Hi'.
      apply Hf in Hi'.
      contradiction (Ha Hi').
    
    destruct (nat_zero n) as [ H0 | H0 ].

    rewrite <- H0 in * |- *.
    assert (a=0) as Ha.
      apply le_S_n in H1;
      auto with arith.
    rewrite Ha in * |- *.
    simpl in Hn.
    apply Nat.max_lub_lt_iff in Hn.
    assert (f 0 = 0) as Hf0.
      destruct Hn as [ Hn _ ].
      apply le_S_n in Hn;
      auto with arith.
    clear n a H1 H2 Hf' H0 Ha Hn.
    split.
    intros a Ha; red in Ha; exists 0.
    apply le_S_n in Ha; rewrite Hf0; auto with arith.
    intros a [ i [ Hi1 Hi2 ] ]. 
    cutrewrite (i=0) in Hi2. 
    red; rewrite Hf0 in Hi2; rewrite <- Hi2; auto.
    apply le_S_n in Hi1; auto with arith.

    apply ltSn_disj in H1.
    destruct H1 as [ H1 | H1 ].

    assert (a <> n) as Han.
      intros H; contradict H1; rewrite H; auto with arith.
    set (g := permut f a n).
    generalize (permut_inj f _ _ Han Hf); intros Hg.
    apply lt_n_Sm_le in Hn.
    assert (forall i, i < n -> g i < n) as Hg1.
      intros i Hi. apply lt_le_trans with (2 := Hn).
      rewrite <- H2.
      unfold g.
      destruct (eq_nat_dec i a) as [ Hia | Hia ].
      rewrite Hia, permut_a. 
      apply Hf'; auto with arith.
      rewrite permut_not_ab; auto with arith.
      intros H; contradict Hi; rewrite H; auto with arith.
    assert (fnat_max g n < n) as Hg'.
      destruct (fnat_max_exists g n) as [ i [ Hi1 Hi2 ] ]; trivial.
      rewrite <- Hi2.
      apply Hg1; trivial.
    generalize (IHn _ Hg Hg'); intros Hg''.
    assert (g n = n) as Hgn.
      apply le_antisym.
      unfold g. 
      rewrite permut_b, H2; auto. 
      destruct (le_or_lt n (g n)) as [ | H3]; trivial.
      generalize (proj1 Hg'' (g n)); intros Hg2.
      apply Hg2 in H3. destruct H3 as [ k [ Hk1 Hk2 ] ].
      apply Hg in Hk2. 
      contradict Hk1; rewrite Hk2; auto with arith.
    generalize (maxf_step n _ Hg H0 Hg' Hgn Hg'').
    intros Hg2. red in Hg2 |- *.
    apply id_transitive with (1 := Hg2), id_symetric, permut_img; auto with arith.

    rewrite H1 in * |- *; clear a H1.
    simpl in H2. symmetry in H2. apply Nat.max_l_iff in H2.
    simpl in Hn. apply Nat.max_lub_lt_iff in Hn. destruct Hn as [ H1 _ ].
    assert (forall i, i < n -> f i < f n) as Hn. 
      intros i Hi; apply Hf'; auto with arith.
      intros H; contradict Hi; rewrite H; auto with arith.
    assert (fnat_max f n < f n) as Hn'.
    destruct (le_lt_or_eq _ _ H2) as [ | H3 ]; trivial.
    destruct (fnat_max_exists f _ H0) as [ i [Hi1 Hi2] ].
    rewrite <- Hi2 in H3. apply Hf in H3. 
    contradict Hi1; rewrite H3; auto with arith.

    assert (fnat_max f n < n) as H3. 
      apply lt_le_trans with (1 := Hn').
      apply le_S_n in H1; auto.
    generalize (IHn _ Hf H3). intros Himg.

    generalize (imgsect_maxfn _ _ Himg). intros Hfn.
    assert (f n = n) as H4.
      apply le_antisym.
      apply le_S_n in H1; auto.
      rewrite <- pred_of_minus in Hfn.
      rewrite Hfn in Hn'.
      unfold lt in Hn'.
      rewrite <- (S_pred _ 0) in Hn'; auto.
    apply (maxf_step _ _ Hf H0 H3 H4 Himg).
  Qed.

  Proposition fnat_max_img f n : fnat_inj f -> fnat_max f n < n -> sect n ~ img f n.
  Proof. apply maxf_rec; auto. Qed.

End Maxf_img.

Proposition fnat_max_lower f : fnat_inj f -> forall n, n-1 <= fnat_max f n.
Proof.
  intros Hf n. 
  destruct (le_lt_dec n (fnat_max f n)) as [ H | H ].
  apply le_trans with (2 := H); apply le_minus.
  destruct (proj1 (fnat_max_img _ _ Hf H) (n-1)) as [ i [ H1 H2 ] ]. 
  red; apply lt_minus; auto.
  apply le_trans with (2 := H); auto with arith.
  rewrite <- H2. 
  apply fnat_max_max; auto.
Qed.

Proposition fnat_inj_prop f : fnat_inj f -> forall n, { k | k <= n <= f k }.
Proof.
  intros Hf n.
  generalize (fnat_max_lower _ Hf (S n)); intros Hn.
  destruct (fnat_max_exists f (S n)) as [ i [ H1 H2 ] ].
  auto with arith.
  exists i.
  apply le_S_n in H1.
  rewrite <- pred_of_minus, <- pred_Sn, <- H2 in Hn.
  auto.
Qed.

Proposition fnat_inj_prop' f : fnat_inj f -> forall n, exists k, k <= n <= f k.
Proof.
  intros Hf n.
  destruct (fnat_inj_prop _ Hf n) as [ x ].
  exists x; auto.
Qed.

Proposition fnat_inj_unbounded f : fnat_inj f -> unbounded (d_image f).
Proof.
  intros Hf n.
  destruct (fnat_inj_prop f Hf n) as [ k Hk ].
  exists (f k).
  split. tauto.
  exists k; auto.
Qed.



